/*
 * 
 * This work is licensed under The MIT License (MIT)
 * 
 * Copyright (c) 2014-2016 Johannes Truschnigg <johannes@truschnigg.info>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *

 * This program is supposed to replace the following ad-hoc script for
 * archiving WAL segments:

 * archive_command = 'test ! -f /wal_archive_destination/%f && cp %p /wal_archive_destination/%f && chmod 644 /wal_archive_destination/%f'

 *
 * Its calling syntax is `<progname> SOURCE DEST`
 *
 * It shall:
 * 0.) fail safely
 * 1.) fail if DEST already exists
 * 2.) fail if copying SOURCE to DEST goes wrong
 * 3.) ensure DEST is world-readable
 * 4.) handle signals as safely as possible
 *
 * It will also report concise status information on stdout or stderr.
 *
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <libgen.h>
#include <limits.h>

#ifndef PROGNAME
#define PROGNAME "pg_archive_wal_segment"
#endif

#ifndef _VERSION
#define _VERSION "devel"
#endif

#ifndef _TIMEOUTSEC
#define _TIMEOUTSEC 30
#endif

#ifndef _BUFSIZE
#define _BUFSIZE 524288
#endif

#define DEATH_NOTE "Caught a fatal signal, performing cleanup.\n"


int fd_src, fd_dst, fd_dstdir;
const char *path_src, *path_dst;
char *path_dstdir;

void croak_safely(int sig_num);
int64_t timespecDiff(struct timespec *timeA_p, struct timespec *timeB_p);


int main(int argc, char* argv[])
{
	char iobuffer[_BUFSIZE];
	size_t blocksize;
	ssize_t bytes_read, bytes_written, bytes_copied;
	struct sigaction sig_handler;
	struct timeval time_start, time_finish, time_total;

	memset(&time_start, 0, sizeof(struct timeval));
	memset(&time_finish, 0, sizeof(struct timeval));
	memset(&time_total, 0, sizeof(struct timeval));
	(void) gettimeofday(&time_start, NULL);

	memset(&sig_handler, 0, sizeof(struct sigaction));
	sig_handler.sa_handler = croak_safely;

	bytes_read = 0;
	bytes_written = 0;
	bytes_copied = 0;
	blocksize = sizeof(iobuffer);
	memset(iobuffer, 0, blocksize);
	(void) umask(0); /* disable umask; we want to set permissions explicitly */

	if(argc == 2 && (strncmp("-V", argv[1], 3) == 0))
	{
		fprintf(stderr, "%s version %s (blocksize=%db, timeout=%ds)\n",
				PROGNAME, _VERSION, _BUFSIZE, _TIMEOUTSEC);
		return EXIT_FAILURE;
	}

	if(argc != 3)
	{
		fprintf(stderr, "Usage:\n");
		fprintf(stderr, "\t %s -V\n", PROGNAME);
		fprintf(stderr, "\t %s SOURCE DEST\n", PROGNAME);
		return EXIT_FAILURE;
	}

	path_src = argv[1];
	path_dst = argv[2];
	path_dstdir = dirname(strndup(argv[2], PATH_MAX -1));

	(void) alarm(_TIMEOUTSEC);

	(void) sigaction(SIGINT, &sig_handler, NULL);
	(void) sigaction(SIGHUP, &sig_handler, NULL);
	(void) sigaction(SIGALRM, &sig_handler, NULL);
	(void) sigaction(SIGTERM, &sig_handler, NULL);

	if((fd_src = open(path_src, O_RDONLY)) == -1)
	{
		fprintf(stderr, "%s: FATAL: error while opening source file:\n", PROGNAME);
		perror(path_src);
		return EXIT_FAILURE;
	}
	if((fd_dstdir = open(path_dstdir, O_DIRECTORY|O_RDONLY)) == -1)
	{
		fprintf(stderr, "%s: FATAL: error while opening destination directory:\n", PROGNAME);
		perror(path_dstdir);
		return EXIT_FAILURE;
	}
	if((fd_dst = open(path_dst, O_WRONLY|O_CREAT|O_EXCL, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)) == -1)
	{
		fprintf(stderr, "%s: FATAL: error while opening destination file:\n", PROGNAME);
		perror(path_dst);
		return EXIT_FAILURE;
	}

	while((bytes_read = read(fd_src, iobuffer, blocksize)) != 0)
	{
		/* fprintf(stderr, "%zu bytes read.\n", bytes_read); */
		if(!(bytes_written = write(fd_dst, iobuffer, bytes_read) == bytes_read))
		{
			fprintf(stderr, "%s: FATAL: error while writing destination file:\n", PROGNAME);
			fprintf(stderr, "%s: short write - total bytes copied: %zu - last iteration: %zu read, %zu written\n",
				PROGNAME, bytes_copied, bytes_read, bytes_written);
			croak_safely(0);
		}
		bytes_copied += (size_t) bytes_read;
	}

	(void) close(fd_src);

	if(fsync(fd_dst) != 0)
	{
		fprintf(stderr, "%s: FATAL: error while flushing destination file data:\n", PROGNAME);
		perror(path_dst);
		croak_safely(0);
	}
	if(close(fd_dst) != 0)
	{
		fprintf(stderr, "%s: FATAL: error while closing destination file:\n", PROGNAME);
		perror(path_dst);
		croak_safely(0);
	}
	if(fsync(fd_dstdir) != 0)
	{
		fprintf(stderr, "%s: FATAL: error while flushing destination directory metadata:\n", PROGNAME);
		perror(path_dst);
		croak_safely(0);
	}
	if(close(fd_dstdir) != 0)
	{
		fprintf(stderr, "%s: FATAL: error while closing destination directory:\n", PROGNAME);
		perror(path_dst);
		croak_safely(0);
	}

	(void) gettimeofday(&time_finish, NULL);
	timersub(&time_finish, &time_start, &time_total);

	fprintf(stdout, "%s: '%s' -> '%s' (%ldb, %ld.%04lds)\n",
		PROGNAME, path_src, path_dst, (long int) bytes_copied,
		(long int) time_total.tv_sec, (long int) time_total.tv_usec);
	return EXIT_SUCCESS;
}


void croak_safely(int sig_num)
{
		if(sig_num == 0)
		{
			/* we're not actually handling a signal here, so I
			 * think printf should be safe... */
			printf("%s: error detected, cleaning up.\n", PROGNAME);
		} else if(sig_num == SIGALRM)
		{
			(void) write(2, PROGNAME, sizeof(PROGNAME));
			(void) write(2, ": timeout exceeded; aborting and cleaning up.\n", 47);
		} else
		{
			(void) write(2, PROGNAME, sizeof(PROGNAME));
			(void) write(2, ": ", 2);
			(void) write(2, DEATH_NOTE, sizeof(DEATH_NOTE));
		}
		(void) close(fd_dst);
		(void) unlink(path_dst);
		exit(EXIT_FAILURE);
}

int64_t timespecDiff(struct timespec *timeA_p, struct timespec *timeB_p)
{
	return ((timeA_p->tv_sec * 1000000000) + timeA_p->tv_nsec) -
	((timeB_p->tv_sec * 1000000000) + timeB_p->tv_nsec);
}
